#!/usr/bin/python
import os, stat, time, subprocess, sys
from collections import OrderedDict

# some utility functions
def safeCall(args):
	res = ""
	for arg in args:
		assert arg.find("'") < 0 # ' is not supported
		if len(res): res += " "
		res += "'"+arg+"'"
	return res

# abstract representation of rules file
class RulesFile:
	def __init__(self):
		self.env = {}
		self.dh = []
		self.rules = OrderedDict()
	
	def env2str(self):
		res = ""
		for name in self.env: # we rely on the name being sane (i.e., no special characters)
			val = self.env[name]
			assert val.find("'") < 0 # ' is not supported
			if len(res): res += " "
			res += name+"='"+val+"'"
		return res

	def write(self, f):
		print >>f, "#!/usr/bin/make -f"
		print >>f, ""
		print >>f, "%:"
		print >>f, '\t'+self.env2str()+' dh $@ '+safeCall(self.dh)
		for rule in self.rules:
			print >>f, ""
			print >>f, "override_dh_"+rule+":"
			for line in self.rules[rule]:
				print >>f, "\t"+line

# build-system specific part of rules file
def cmakeRules(config):
	buildDir = config.get('buildDir', 'build.dir') # "build" is not a good idea, as that's also the name of a target...
	srcDir = os.getcwd()
	r = RulesFile()
	r.dh += ["--buildsystem=cmake", "--builddirectory="+buildDir] # dh parameters
	r.rules['auto_configure'] = [
		safeCall(["mkdir", "-p", buildDir]),
		safeCall(["cd", buildDir]) + " && " +
		  safeCall(["cmake", srcDir, "-DCMAKE_INSTALL_PREFIX=/usr"] + config.get('cmakeParameters', []))
	]
	r.rules['auto_clean'] = [safeCall(['rm', '-f', os.path.join(buildDir, 'CMakeCache.txt')])] # clean old cmake cache
	return r

def automakeRules(config):
	r = RulesFile()
	r.dh += ["--buildsystem=autoconf"]
	r.rules['auto_configure'] = [
		'BUILD_TYPE=$$(dpkg-architecture -qDEB_BUILD_GNU_TYPE) && ' + # doing the expansion beforehand ensures that we cancel if it fails
		'MULTIARCH=$$(dpkg-architecture -qDEB_BUILD_MULTIARCH) && '+
			'./configure --build=$$BUILD_TYPE '+ # do the escape manually here so we can use the variables (there's no user-controlled string in here anyway)
			'--prefix=/usr --includedir=/usr/include --mandir=/usr/share/man --infodir=/usr/share/info '+
			'--libdir=/usr/lib/$$MULTIARCH --libexecdir=/usr/lib/$$MULTIARCH '+
			'--sysconfdir=/etc --localstatedir=/var '+
			safeCall(config.get('automakeParameters', []))
	]
	r.rules['auto_clean'] = ['rm -f config.status'] # do not re-use old configuration (no need for escaping here, obviously)
	return r

# utility functions
def commandInBuildEnv(config, command):
	schroot = config.get('schroot')
	if schroot is not None: command = ['schroot', '-c', schroot, '--'] + command
	return command

def getArchitecture(config):
	cmd = commandInBuildEnv(config, ['dpkg-architecture', '-qDEB_BUILD_ARCH'])
	p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
	res = p.communicate()[0] # get only stdout
	if p.returncode != 0: raise Exception("Querying dpkg for the architecture failed")
	return res[0:len(res)-1] # chop of the \n at the end

def writeDependency(f, name, list):
	if len(list):
		print >>f, name+": "+', '.join(list)

# actual work functions
def createDebianFiles(config):
	sourceName = config['sourceName']
	binaryName = config.get('binaryName', sourceName+'-local')
	name = config.get('name', os.getlogin())
	email = config.get('email', os.getlogin()+'@'+os.uname()[1]) # user@hostname
	debDir = os.path.expanduser(config['debDir'])
	buildSystem = config['buildSystem']
	version = config['version']
	dbgPackage = config.get('dbgPackage', False)
	packageArchitecture = config.get('architecture', 'any')
	# we return the list of files generated, so we need to know the architecture
	arch = getArchitecture(config)
	files = []
	# create folders
	if not os.path.exists('debian'): os.mkdir('debian')
	if not os.path.exists('debian/source'): os.mkdir('debian/source')
	if not os.path.exists(debDir): os.mkdir(debDir)
	# source format file
	with open('debian/source/format', 'w') as f:
		print >>f, "3.0 (native)"
	# compat file
	with open('debian/compat', 'w') as f:
		print >>f, "9"
	# copyright file
	with open('debian/copyright', 'w') as f:
		print >>f, "Auto-generated by auto-debuild, not suited for distribution"
	# changelog file
	with open('debian/changelog', 'w') as f:
		print >>f, sourceName,"("+version+")","UNRELEASED; urgency=low"
		print >>f, ""
		print >>f, "  * Auto-generated by auto-debuild"
		print >>f, ""
		print >>f, " --",name,"<"+email+">  "+time.strftime('%a, %d %b %Y %H:%M:%S %z')
	# control file
	with open('debian/control', 'w') as f:
		print >>f, "Source:",sourceName
		print >>f, "Section:",config.get('section', 'misc')
		print >>f, "Priority: extra"
		print >>f, "Maintainer: %s <%s>" % (name, email)
		writeDependency(f, 'Build-Depends', ["debhelper (>= 9)"] + config.get('buildDepends', []))
		print >>f, "Standards-Version: 3.9.3"
		print >>f, ""
		print >>f, "Package:",binaryName
		print >>f, "Architecture:",packageArchitecture
		writeDependency(f, "Pre-Depends", ["${misc:Pre-Depends}"] + config.get('binaryPreDepends', []))
		writeDependency(f, "Depends", ["${shlibs:Depends}", "${misc:Depends}"] + config.get('binaryDepends', []))
		writeDependency(f, "Recommends", config.get('binaryRecommends', []))
		writeDependency(f, "Provides", config.get('binaryProvides', [sourceName]))
		print >>f, "Description:",sourceName,"(auto-debuild)"
		print >>f, " Package auto-generated by auto-debuild."
		files.append(os.path.join(debDir, "%s_%s_%s.deb" % (binaryName, version, arch)))
		if dbgPackage:
			print >>f, ""
			print >>f, "Package:",binaryName+"-dbg"
			print >>f, "Architecture:",packageArchitecture
			writeDependency(f, "Depends", ["${misc:Depends}", binaryName+" (= ${binary:Version})"])
			print >>f, "Description:",sourceName,"debug smbols (auto-debuild)"
			print >>f, " Package containing debug symbols for "+sourceName+", auto-generated by auto-debuild."
			files.append(os.path.join(debDir, "%s-dbg_%s_%s.deb" % (binaryName, version, arch)))
	# rules file: build system specific
	with open('debian/rules', 'w') as f:
		# get rule file for build system: may only touch auto_config and auto_clean rules and the dh options
		if buildSystem == 'cmake':
			r = cmakeRules(config)
		elif buildSystem == 'automake':
			r = automakeRules(config)
		else:
			raise Exception("Invalid build system "+buildSystem)
		# global rules
		r.env["DEB_BUILD_OPTIONS"] = 'parallel=2'
		if not dbgPackage:
			# disable debug information
			r.env["DEB_CFLAGS_APPEND"] = '-g0'
			r.env["DEB_CXXFLAGS_APPEND"] = '-g0'
		r.dh += ['--parallel']
		r.rules['builddeb'] = [safeCall(['dh_builddeb', "--destdir="+debDir])] # passing this gobally to dh results in weird problems (like stuff being installed there, and not in the package...)
		r.rules['auto_test'] = []
		r.rules['auto_install'] = [safeCall(['dh_auto_install', '--destdir=debian/'+binaryName])] # install everything into the binary package
		# for debug packages
		if dbgPackage:
			r.rules['strip'] = [safeCall(['dh_strip', '--dbg-package='+binaryName+"-dbg"])] # put debug files in appropriate package
			r.rules['installdocs'] = [safeCall(['dh_installdocs', '--link-doc='+binaryName])] # make the doc folder of the dbg package a symlink
		# dump it to a file
		r.write(f)
	mode = os.stat('debian/rules').st_mode
	os.chmod('debian/rules', mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
	# return list of files affected
	return files

def buildDebianPackage(config):
	commands = ['dpkg-checkbuilddeps', 'debian/rules clean', 'debian/rules build', 'fakeroot debian/rules binary', 'debian/rules clean']
	command = ['nice', 'bash', '-c', ' && '.join(commands)]
	subprocess.check_call(commandInBuildEnv(config, command))

# all at once
def createAndInstall(config, overwriteCheck = False):
	# generate debian files
	files = createDebianFiles(config)
	# check if a file is overwritten
	if overwriteCheck:
		for file in files:
			if os.path.exists(file):
				if raw_input("Do you want to overwrite %s (y/N)? " % file).lower() != "y":
					sys.exit(1)
	# run compilation
	buildDebianPackage(config)
	# install files
	print "Installing created deb files..."
	subprocess.check_call(['sudo', 'dpkg', '--install'] + files)

# if we are called directly as script
if __name__ == "__main__":
	# get config
	import imp
	config = imp.load_source('config', 'debian/auto-debuild.conf').__dict__
	os.remove('debian/auto-debuild.confc')
	# and go for it
	createAndInstall(config, overwriteCheck=True)
